# Generated by Django 5.2 on 2025-07-06 08:47

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("lamindb", "0113_lower_case_branch_and_space_names"),
    ]

    operations = [
        migrations.AlterField(
            model_name="run",
            name="_status_code",
            field=models.SmallIntegerField(db_index=True, default=None, null=True),
        ),
        migrations.RunSQL(
            sql="""
            UPDATE lamindb_run
            SET _status_code = NULL
            WHERE _status_code = 0;
            """,
        ),
    ]
