# Generated by Django 5.2 on 2025-06-03 19:37

from django.db import connection, migrations


class Migration(migrations.Migration):
    dependencies = [
        ("lamindb", "0105_record_unique_name"),
    ]

    operations = [
        migrations.RunSQL(
            sql="""
                UPDATE lamindb_transform
                SET key = '__lamindb_transfer__/' || SUBSTR(key, 11)
                WHERE key LIKE 'transfers/%';
            """
            if connection.vendor == "sqlite"
            else """
                UPDATE lamindb_transform
                SET key = CONCAT('__lamindb_transfer__/', SUBSTRING(key FROM 11))
                WHERE key LIKE 'transfers/%';
            """
        ),
    ]
