# Generated by Django 5.2 on 2025-05-06 20:34

import django.db.models.deletion
from django.db import migrations

import lamindb.base.fields


class Migration(migrations.Migration):
    dependencies = [
        ("lamindb", "0091_alter_featurevalue_options_alter_space_options_and_more"),
    ]

    operations = [
        migrations.AlterField(
            model_name="artifactfeaturevalue",
            name="artifact",
            field=lamindb.base.fields.ForeignKey(
                blank=True,
                on_delete=django.db.models.deletion.CASCADE,
                related_name="links_featurevalue",
                to="lamindb.artifact",
            ),
        ),
        migrations.AlterField(
            model_name="artifactfeaturevalue",
            name="featurevalue",
            field=lamindb.base.fields.ForeignKey(
                blank=True,
                on_delete=django.db.models.deletion.PROTECT,
                related_name="links_artifact",
                to="lamindb.featurevalue",
            ),
        ),
        migrations.AlterField(
            model_name="artifactparamvalue",
            name="artifact",
            field=lamindb.base.fields.ForeignKey(
                blank=True,
                on_delete=django.db.models.deletion.CASCADE,
                related_name="links_paramvalue",
                to="lamindb.artifact",
            ),
        ),
        migrations.AlterField(
            model_name="artifactparamvalue",
            name="paramvalue",
            field=lamindb.base.fields.ForeignKey(
                blank=True,
                on_delete=django.db.models.deletion.PROTECT,
                related_name="links_artifact",
                to="lamindb.paramvalue",
            ),
        ),
        migrations.AlterField(
            model_name="runparamvalue",
            name="paramvalue",
            field=lamindb.base.fields.ForeignKey(
                blank=True,
                on_delete=django.db.models.deletion.PROTECT,
                related_name="links_run",
                to="lamindb.paramvalue",
            ),
        ),
        migrations.AlterField(
            model_name="runparamvalue",
            name="run",
            field=lamindb.base.fields.ForeignKey(
                blank=True,
                on_delete=django.db.models.deletion.CASCADE,
                related_name="links_paramvalue",
                to="lamindb.run",
            ),
        ),
    ]
