# Generated by Django 5.2 on 2025-01-27 07:22

import django.db.models.deletion
from django.db import migrations

import lamindb.base.fields


class Migration(migrations.Migration):
    dependencies = [
        ("lamindb", "0083_alter_feature_is_type_alter_flextable_is_type_and_more"),
    ]

    operations = [
        migrations.AlterField(
            model_name="schemafeature",
            name="feature",
            field=lamindb.base.fields.ForeignKey(
                blank=True,
                on_delete=django.db.models.deletion.PROTECT,
                related_name="links_schema",
                to="lamindb.feature",
            ),
        ),
        migrations.AlterField(
            model_name="schemafeature",
            name="schema",
            field=lamindb.base.fields.ForeignKey(
                blank=True,
                on_delete=django.db.models.deletion.CASCADE,
                related_name="links_feature",
                to="lamindb.schema",
            ),
        ),
    ]
