# Generated by Django 5.2 on 2025-07-04 20:43

import django.db.models.deletion
from django.db import migrations

import lamindb.base.fields


class Migration(migrations.Migration):
    dependencies = [
        ("lamindb", "0111_remove_record__sort_order"),
    ]

    operations = [
        migrations.AlterField(
            model_name="recordartifact",
            name="feature",
            field=lamindb.base.fields.ForeignKey(
                blank=True,
                on_delete=django.db.models.deletion.PROTECT,
                related_name="links_recordartifact",
                to="lamindb.feature",
            ),
        ),
        migrations.AlterField(
            model_name="recordjson",
            name="feature",
            field=lamindb.base.fields.ForeignKey(
                blank=True,
                on_delete=django.db.models.deletion.PROTECT,
                related_name="links_recordjson",
                to="lamindb.feature",
            ),
        ),
        migrations.AlterField(
            model_name="recordrecord",
            name="feature",
            field=lamindb.base.fields.ForeignKey(
                blank=True,
                on_delete=django.db.models.deletion.PROTECT,
                related_name="links_recordrecord",
                to="lamindb.feature",
            ),
        ),
        migrations.AlterField(
            model_name="recordrun",
            name="feature",
            field=lamindb.base.fields.ForeignKey(
                blank=True,
                on_delete=django.db.models.deletion.PROTECT,
                related_name="links_recordrun",
                to="lamindb.feature",
            ),
        ),
        migrations.AlterField(
            model_name="recordulabel",
            name="feature",
            field=lamindb.base.fields.ForeignKey(
                blank=True,
                on_delete=django.db.models.deletion.PROTECT,
                related_name="links_recordulabel",
                to="lamindb.feature",
            ),
        ),
        migrations.AlterField(
            model_name="recordperson",
            name="feature",
            field=lamindb.base.fields.ForeignKey(
                blank=True,
                on_delete=django.db.models.deletion.PROTECT,
                related_name="links_recordperson",
                to="lamindb.feature",
            ),
        ),
        migrations.AlterField(
            model_name="recordproject",
            name="feature",
            field=lamindb.base.fields.ForeignKey(
                blank=True,
                on_delete=django.db.models.deletion.PROTECT,
                related_name="links_recordproject",
                to="lamindb.feature",
            ),
        ),
        migrations.AlterField(
            model_name="recordreference",
            name="feature",
            field=lamindb.base.fields.ForeignKey(
                blank=True,
                on_delete=django.db.models.deletion.PROTECT,
                related_name="links_recordreference",
                to="lamindb.feature",
            ),
        ),
        migrations.AlterField(
            model_name="recordartifact",
            name="value",
            field=lamindb.base.fields.ForeignKey(
                blank=True,
                on_delete=django.db.models.deletion.PROTECT,
                related_name="links_in_record",
                to="lamindb.artifact",
            ),
        ),
    ]
