# Generated by Django 5.2 on 2025-10-13 07:42

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("lamindb", "0133_artifactuser_artifact_users"),
    ]

    operations = [
        migrations.AddField(
            model_name="run",
            name="params",
            field=models.JSONField(null=True),
        ),
    ]
