# Generated by Django 5.1.12 on 2025-10-02 15:17

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("lamindb", "0130_branch_space_alter_artifactblock_artifact_and_more"),
    ]

    operations = [
        migrations.AddConstraint(
            model_name="record",
            constraint=models.UniqueConstraint(
                fields=("name", "type", "space"), name="unique_name_type_space"
            ),
        ),
    ]
