# Generated by Django 5.2 on 2025-02-06 07:10

from django.db import migrations, models

import lamindb.base.fields


class Migration(migrations.Migration):
    dependencies = [
        ("lamindb", "0085_alter_feature_is_type_alter_flextable_is_type_and_more"),
    ]

    operations = [
        migrations.AlterField(
            model_name="transform",
            name="hash",
            field=lamindb.base.fields.CharField(
                blank=True,
                db_index=True,
                default=None,
                max_length=22,
                null=True,
                unique=True,
            ),
        ),
        migrations.AlterField(
            model_name="artifact",
            name="hash",
            field=lamindb.base.fields.CharField(
                blank=True,
                db_index=True,
                default=None,
                max_length=22,
                null=True,
                unique=True,
            ),
        ),
        migrations.AlterField(
            model_name="collection",
            name="hash",
            field=lamindb.base.fields.CharField(
                blank=True,
                db_index=True,
                default=None,
                max_length=22,
                null=True,
                unique=True,
            ),
        ),
        migrations.CreateModel(
            name="Migration",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "app",
                    lamindb.base.fields.CharField(
                        blank=True, default=None, max_length=255
                    ),
                ),
                (
                    "name",
                    lamindb.base.fields.CharField(
                        blank=True, default=None, max_length=255
                    ),
                ),
                ("applied", lamindb.base.fields.DateTimeField(blank=True)),
            ],
            options={
                "db_table": "django_migrations",
                "managed": False,
            },
        ),
        migrations.AlterField(
            model_name="param",
            name="dtype",
            field=lamindb.base.fields.CharField(
                blank=True, db_index=True, default=None, max_length=64, null=True
            ),
        ),
        migrations.AlterField(
            model_name="param",
            name="dtype",
            field=lamindb.base.fields.CharField(
                blank=True, db_index=True, default=None, max_length=255, null=True
            ),
        ),
    ]
