# Generated by Django 5.2 on 2025-06-30 12:54

from django.db import migrations


class Migration(migrations.Migration):
    dependencies = [
        ("lamindb", "0107_add_schema_to_record"),
    ]

    operations = [
        migrations.RemoveField(
            model_name="record",
            name="sheet",
        ),
        migrations.RemoveField(
            model_name="project",
            name="sheets",
        ),
        migrations.AlterUniqueTogether(
            name="sheetproject",
            unique_together=None,
        ),
        migrations.DeleteModel(
            name="Sheet",
        ),
        migrations.DeleteModel(
            name="SheetProject",
        ),
    ]
