# Generated by Django 5.2 on 2025-07-26 15:55

from django.db import migrations, models

import lamindb.base.fields


class Migration(migrations.Migration):
    dependencies = [
        ("lamindb", "0115_alter_space_uid"),
    ]

    operations = [
        migrations.RemoveConstraint(
            model_name="artifact",
            name="unique_artifact_storage_key_hash",
        ),
        migrations.AlterField(
            model_name="record",
            name="description",
            field=lamindb.base.fields.CharField(
                blank=True, db_index=True, default=None, max_length=255, null=True
            ),
        ),
        migrations.AlterField(
            model_name="reference",
            name="text",
            field=lamindb.base.fields.TextField(
                blank=True, db_index=True, default=None, null=True
            ),
        ),
        migrations.AlterField(
            model_name="reference",
            name="url",
            field=lamindb.base.fields.URLField(blank=True, db_index=True, null=True),
        ),
        migrations.AlterField(
            model_name="run",
            name="name",
            field=lamindb.base.fields.CharField(
                blank=True, db_index=True, default=None, max_length=150, null=True
            ),
        ),
        migrations.AddConstraint(
            model_name="artifact",
            constraint=models.UniqueConstraint(
                fields=("storage", "key", "hash"),
                name="unique_artifact_storage_key_hash",
            ),
        ),
    ]
