# Generated by Django 5.1.12 on 2025-09-28 23:37

from django.db import migrations


class Migration(migrations.Migration):
    dependencies = [
        (
            "lamindb",
            "0123_alter_artifact_description_alter_branch_description_and_more",
        ),
    ]

    # this migration was later reverted, so we do not need to run it
    operations = []  # type: ignore
