# Generated by Django 5.1.12 on 2025-10-04 10:57

import django.db.models.deletion
from django.db import migrations, models

import lamindb.base.fields


class Migration(migrations.Migration):
    dependencies = [
        ("lamindb", "0131_record_unique_name_type_space"),
    ]

    operations = [
        migrations.AddField(
            model_name="record",
            name="parents",
            field=models.ManyToManyField(related_name="children", to="lamindb.record"),
        ),
        migrations.AddField(
            model_name="record",
            name="reference",
            field=lamindb.base.fields.CharField(
                blank=True, db_index=True, default=None, max_length=255, null=True
            ),
        ),
        migrations.AddField(
            model_name="record",
            name="reference_type",
            field=lamindb.base.fields.CharField(
                blank=True, db_index=True, default=None, max_length=25, null=True
            ),
        ),
        migrations.RenameField(
            model_name="record",
            old_name="ulabels",
            new_name="linked_ulabels",
        ),
        migrations.AlterField(
            model_name="record",
            name="linked_ulabels",
            field=models.ManyToManyField(
                related_name="linked_in_records",
                through="lamindb.RecordULabel",
                to="lamindb.ulabel",
            ),
        ),
        migrations.AlterField(
            model_name="record",
            name="run",
            field=lamindb.base.fields.ForeignKey(
                blank=True,
                default=lamindb.models.run.current_run,
                editable=False,
                null=True,
                on_delete=django.db.models.deletion.PROTECT,
                related_name="output_records",
                to="lamindb.run",
            ),
        ),
    ]
