"""Base library.

Is available also when no instance is setup.

Modules:

.. autosummary::
   :toctree: .

   uids
   types
   fields
   dtypes

Utils:

.. autosummary::
   :toctree: .

   doc_args
   deprecated

"""

from lamindb_setup.core import deprecated, doc_args

from . import dtypes, fields, types, uids
