# Generated by Django 5.2 on 2025-10-15 09:35

from django.db import migrations, models

import lamindb.base.fields


class Migration(migrations.Migration):
    dependencies = [
        ("lamindb", "0134_run_params"),
    ]

    operations = [
        migrations.AlterField(
            model_name="collection",
            name="hash",
            field=lamindb.base.fields.CharField(
                blank=True, db_index=True, default=None, max_length=22, null=True
            ),
        ),
        migrations.AddConstraint(
            model_name="collection",
            constraint=models.UniqueConstraint(
                fields=("key", "hash"), name="unique_collection_key_hash_not_null"
            ),
        ),
    ]
