# Generated by Django 5.2 on 2025-06-03 19:37

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("lamindb", "0104_alter_branch_uid"),
    ]

    operations = [
        migrations.AddConstraint(
            model_name="record",
            constraint=models.UniqueConstraint(
                condition=models.Q(("is_type", True)),
                fields=("name",),
                name="unique_name",
            ),
        ),
    ]
