# Generated by Django 5.1.12 on 2025-10-02 10:27

import django.db.models.deletion
from django.db import migrations, models

import lamindb.base.fields


class Migration(migrations.Migration):
    dependencies = [
        ("lamindb", "0129_remove_feature_page_remove_project_page_and_more"),
    ]

    operations = [
        migrations.AddField(
            model_name="branch",
            name="space",
            field=lamindb.base.fields.ForeignKey(
                blank=True,
                db_default=1,
                default=1,
                on_delete=django.db.models.deletion.PROTECT,
                related_name="+",
                to="lamindb.space",
            ),
        ),
        migrations.AlterField(
            model_name="artifactblock",
            name="artifact",
            field=models.ForeignKey(
                default=1,
                on_delete=django.db.models.deletion.CASCADE,
                related_name="blocks",
                to="lamindb.artifact",
            ),
            preserve_default=False,
        ),
        migrations.AlterField(
            model_name="artifactblock",
            name="vertical_pos",
            field=models.FloatField(db_default=0, db_index=True, default=0),
        ),
        migrations.AlterField(
            model_name="branchblock",
            name="branch",
            field=models.ForeignKey(
                default=1,
                on_delete=django.db.models.deletion.CASCADE,
                related_name="blocks",
                to="lamindb.branch",
            ),
            preserve_default=False,
        ),
        migrations.AlterField(
            model_name="branchblock",
            name="vertical_pos",
            field=models.FloatField(db_default=0, db_index=True, default=0),
        ),
        migrations.AlterField(
            model_name="collectionblock",
            name="vertical_pos",
            field=models.FloatField(db_default=0, db_index=True, default=0),
        ),
        migrations.AlterField(
            model_name="featureblock",
            name="feature",
            field=models.ForeignKey(
                default=1,
                on_delete=django.db.models.deletion.CASCADE,
                related_name="blocks",
                to="lamindb.feature",
            ),
            preserve_default=False,
        ),
        migrations.AlterField(
            model_name="featureblock",
            name="vertical_pos",
            field=models.FloatField(db_default=0, db_index=True, default=0),
        ),
        migrations.AlterField(
            model_name="projectblock",
            name="project",
            field=models.ForeignKey(
                default=1,
                on_delete=django.db.models.deletion.CASCADE,
                related_name="blocks",
                to="lamindb.project",
            ),
            preserve_default=False,
        ),
        migrations.AlterField(
            model_name="projectblock",
            name="vertical_pos",
            field=models.FloatField(db_default=0, db_index=True, default=0),
        ),
        migrations.AlterField(
            model_name="recordblock",
            name="record",
            field=models.ForeignKey(
                default=1,
                on_delete=django.db.models.deletion.CASCADE,
                related_name="blocks",
                to="lamindb.record",
            ),
            preserve_default=False,
        ),
        migrations.AlterField(
            model_name="recordblock",
            name="vertical_pos",
            field=models.FloatField(db_default=0, db_index=True, default=0),
        ),
        migrations.AlterField(
            model_name="rootblock",
            name="vertical_pos",
            field=models.FloatField(db_default=0, db_index=True, default=0),
        ),
        migrations.AlterField(
            model_name="runblock",
            name="run",
            field=models.ForeignKey(
                default=1,
                on_delete=django.db.models.deletion.CASCADE,
                related_name="blocks",
                to="lamindb.run",
            ),
            preserve_default=False,
        ),
        migrations.AlterField(
            model_name="runblock",
            name="vertical_pos",
            field=models.FloatField(db_default=0, db_index=True, default=0),
        ),
        migrations.AlterField(
            model_name="schemablock",
            name="schema",
            field=models.ForeignKey(
                default=1,
                on_delete=django.db.models.deletion.CASCADE,
                related_name="blocks",
                to="lamindb.schema",
            ),
            preserve_default=False,
        ),
        migrations.AlterField(
            model_name="schemablock",
            name="vertical_pos",
            field=models.FloatField(db_default=0, db_index=True, default=0),
        ),
        migrations.AlterField(
            model_name="spaceblock",
            name="space",
            field=models.ForeignKey(
                default=1,
                on_delete=django.db.models.deletion.CASCADE,
                related_name="blocks",
                to="lamindb.space",
            ),
            preserve_default=False,
        ),
        migrations.AlterField(
            model_name="spaceblock",
            name="vertical_pos",
            field=models.FloatField(db_default=0, db_index=True, default=0),
        ),
        migrations.AlterField(
            model_name="transformblock",
            name="vertical_pos",
            field=models.FloatField(db_default=0, db_index=True, default=0),
        ),
    ]
