# Generated by Django 5.1.12 on 2025-09-30 17:16

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        (
            "lamindb",
            "0126_alter_artifact_is_locked_alter_collection_is_locked_and_more",
        ),
    ]

    operations = [
        migrations.AlterField(
            model_name="run",
            name="_status_code",
            field=models.SmallIntegerField(
                db_default=-3, db_index=True, default=-3, null=True
            ),
        ),
        migrations.AddConstraint(
            model_name="feature",
            constraint=models.CheckConstraint(
                condition=models.Q(
                    ("is_type", True), ("dtype__isnull", False), _connector="OR"
                ),
                name="dtype_not_null_when_is_type_false",
            ),
        ),
    ]
