# Generated by Django 5.2 on 2025-10-01 09:07

from django.db import migrations

import lamindb.base.fields


class Migration(migrations.Migration):
    dependencies = [
        ("lamindb", "0127_alter_run_status_code_feature_dtype"),
    ]

    operations = [
        migrations.AddField(
            model_name="artifact",
            name="_real_key",
            field=lamindb.base.fields.CharField(
                blank=True, db_index=True, default=None, max_length=1024, null=True
            ),
        ),
    ]
