# Generated by Django 5.1.12 on 2025-09-29 00:46

from django.db import migrations

import lamindb.base.fields


class Migration(migrations.Migration):
    dependencies = [
        ("lamindb", "0124_page_artifact_page_collection_page_feature_page_and_more"),
    ]

    operations = [
        migrations.AddField(
            model_name="artifact",
            name="is_locked",
            field=lamindb.base.fields.BooleanField(blank=True, default=False),
        ),
        migrations.AddField(
            model_name="collection",
            name="is_locked",
            field=lamindb.base.fields.BooleanField(blank=True, default=False),
        ),
        migrations.AddField(
            model_name="feature",
            name="is_locked",
            field=lamindb.base.fields.BooleanField(blank=True, default=False),
        ),
        migrations.AddField(
            model_name="featurevalue",
            name="is_locked",
            field=lamindb.base.fields.BooleanField(blank=True, default=False),
        ),
        migrations.AddField(
            model_name="project",
            name="is_locked",
            field=lamindb.base.fields.BooleanField(blank=True, default=False),
        ),
        migrations.AddField(
            model_name="record",
            name="is_locked",
            field=lamindb.base.fields.BooleanField(blank=True, default=False),
        ),
        migrations.AddField(
            model_name="recordrecord",
            name="is_locked",
            field=lamindb.base.fields.BooleanField(blank=True, default=False),
        ),
        migrations.AddField(
            model_name="reference",
            name="is_locked",
            field=lamindb.base.fields.BooleanField(blank=True, default=False),
        ),
        migrations.AddField(
            model_name="run",
            name="is_locked",
            field=lamindb.base.fields.BooleanField(blank=True, default=False),
        ),
        migrations.AddField(
            model_name="schema",
            name="is_locked",
            field=lamindb.base.fields.BooleanField(blank=True, default=False),
        ),
        migrations.AddField(
            model_name="storage",
            name="is_locked",
            field=lamindb.base.fields.BooleanField(blank=True, default=False),
        ),
        migrations.AddField(
            model_name="transform",
            name="is_locked",
            field=lamindb.base.fields.BooleanField(blank=True, default=False),
        ),
        migrations.AddField(
            model_name="ulabel",
            name="is_locked",
            field=lamindb.base.fields.BooleanField(blank=True, default=False),
        ),
    ]
